"""
pdfparse - extract useful textual data out of compressed pdf files

This only works with gzip-compressed pdf files at the moment, because that's
all I've seen online.  Patches to support more are welcome.
"""

__author__ = "Tim Hatch <tim@timhatch.com>"
__revision__ = "$Rev: $"[6:-2]
__date__  = "$Date: $"[7:-2]
__license__ = "BSD"
__all__ = ["get_streams", "Stream"]


import re
import sys

from zlib import decompress

class PdfError(Exception): pass

r_stream = re.compile(r"stream\r?\n([Hx][\x89\x9c].+?)endstream", re.DOTALL)

r_row = re.compile(r"\[(?P<cells>.+)\]TJ")
r_cell = re.compile(r"(?:\(((?:\\.|[^)])*)\)|(-[0-9.]*))")
r_esc = re.compile(r"(\\.|[^)])")
r_textobj = re.compile(r"BT\s+(\d+) (\d+) (\d+) (\d+) (?P<x>[0-9.]+) (?P<y>[0-9.]+) Tm\s+\((?P<text>(?:\\.|[^)])*)\)\s+Tj\s+ET")

def removebackslashes(s):
    return r_esc.sub(fixbs, s)

def fixbs(m):
    """
    re.sub callback for removing backslashes
    """
    s = m.group(1)
    if s[0] == "\\":
        return s[1]
    else:
        return s

def parse_row(r):
    r"""
    Helper to estract rows from an uncompressed pdf stream and generate
    a list of cells.

    >>> r1 = '(7-11)-31335.9(1629 W. UNIVERSITY)-5985(12-15-05)-2406.7(97)'
    >>> r2 = "(ALBERTSON'S \(STORE\))-22033.6(2434 I-35E SOUTH)-7317.5(07-25-05)-2406.7(97)"
    >>> parse_row(r1)
    ['7-11', '1629 W. UNIVERSITY', '12-15-05', '97']
    >>> parse_row(r2)
    ["ALBERTSON'S (STORE)", '2434 I-35E SOUTH', '07-25-05', '97']
    
    Remember to verify that the regexps work:
    
    >>> list(m.groups() for m in r_cell.finditer(r1))
    [('7-11', None), (None, '-31335.9'), ('1629 W. UNIVERSITY', None), (None, '-5985'), ('12-15-05', None), (None, '-2406.7'), ('97', None)]
    >>> list(m.groups() for m in r_cell.finditer(r2))
    [("ALBERTSON'S \\(STORE\\)", None), (None, '-22033.6'), ('2434 I-35E SOUTH', None), (None, '-7317.5'), ('07-25-05', None), (None, '-2406.7'), ('97', None)]
    """
    
    c = []
    for m in r_cell.finditer(r):
        if m.group(1):
            c.append(removebackslashes(m.group(1)))
    return c
        

class TextObj(object):
    def __init__(self, x, y, text):
        self.x = float(x)
        self.y = float(y)
        self.text = text
    def __str__(self):
        return self.text
    def __repr__(self):
        return "TextObj(%r, %r, %r)" % (self.x, self.y, self.text)

class Stream(object):
    """
    The only useful member is 'rows' for now.  Once more generic support is
    added, they will be made public here too.
    
    >>> data = list(list(get_streams(file('comprehensive.pdf', 'rb')))[0].rows())
    >>> data[0]
    ['ESTABLISHMENT NAME', 'PHYSICAL ADDRESS', 'DATE DONE', 'SCORE']
    >>> data[4]
    ['7-11', '1115 E. UNIVERSITY', '01-26-05', '100']
    
    """
    def __init__(self, data, n):
        self.data = data
        self.n = n # Index within file, of text-looking streams only
    def rows(self):
        for m in r_row.finditer(self.data):
            yield parse_row(m.group(1))
    def textobjs(self):
        for m in r_textobj.finditer(self.data):
            yield TextObj(m.group('x'), m.group('y'), removebackslashes(m.group('text')))

def get_streams(fo):
    """
    Extract streams from a file-like object and decompress them (generating
    Stream objects for each).  In my testing, one Stream exists per page.
    """
    data = fo.read()
    i = 0
    for d in r_stream.finditer(data):
        dec = decompress(d.group(1))
        if dec[0] != "/" and dec[0] != "q":
            raise PdfError("Invalid chunk %d" % i)
        yield Stream(dec, i)
        i += 1
        

def export_streams(fo, prefix='out'):
    """
    Debug function to dump data to a file-like object.  Useful for making sure
    the chunks you're interested in are indeed being decompressed.
    """
    data = fo.read()
    i = 0
    for m in r_stream.finditer(data):
        dec = decompress(m.group(1))
        if dec[0] != "/" and dec[0] != "q":
            raise PdfError("Invalid chunk %d" % i)
        f = file("%s.%d" % (prefix, i), "wb")
        f.write(dec)
        f.close()
        i += 1

def _test():
    from doctest import testmod
    testmod()

if __name__ == "__main__":
    if "-t" in sys.argv:
        _test()
    else:
        if len(sys.argv) > 1:
            files = sys.argv[1:]
        else:
            files = ['comprehensive.pdf']
        
        for f in files:
            print "Examining", f
            from pprint import pprint
            for s in get_streams(file(f, 'rb')):
                print "  New stream"
                pprint(list(s.rows()))
